from sampo.schemas.apply_queue import ApplyQueue
from sampo.schemas.utils import uuid_str
from sampo.schemas.contractor import Contractor
from sampo.schemas.exceptions import NoSufficientContractorError, NotEnoughMaterialsInDepots, NoAvailableResources
from sampo.schemas.graph import WorkGraph, GraphNode, EdgeType
from sampo.schemas.identifiable import Identifiable
from sampo.schemas.interval import Interval, IntervalGaussian, IntervalUniform
from sampo.schemas.landscape import LandscapeConfiguration, ResourceHolder, ResourceSupply, Road, MaterialDelivery
from sampo.schemas.project import ScheduledProject
from sampo.schemas.requirements import BaseReq, MaterialReq, ZoneReq, EquipmentReq, WorkerReq, ConstructionObjectReq
from sampo.schemas.resources import Resource, Worker, WorkerProductivityMode, ConstructionObject, Material, Equipment
from sampo.schemas.schedule import Schedule
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.serializable import JSONSerializable, AutoJSONSerializable
from sampo.schemas.sorted_list import ExtendedSortedList
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, WorkEstimationMode
from sampo.schemas.types import ScheduleEvent, EventType
from sampo.schemas.works import WorkUnit
from sampo.schemas.zones import Zone, ZoneTransition, ZoneStatuses, ZoneConfiguration, DefaultZoneStatuses



from sampo.utilities.visualization.resources import resource_employment_fig
from sampo.utilities.visualization.schedule import schedule_gant_chart_fig
from sampo.utilities.visualization.work_graph import work_graph_fig



from dataclasses import dataclass
from typing import Optional


@dataclass
class Identifiable:
    """
    A base class for all unique entities

    :param id: unique id for the object
    :param name: name of for the object
    """
    id: str
    name: Optional[str]

    def __hash__(self):
        return hash(self.id)



from abc import ABC, abstractmethod
from enum import Enum
from operator import attrgetter
from random import Random
from typing import Optional

import numpy.random

from sampo.schemas.requirements import WorkerReq
from sampo.schemas.resources import Worker
from sampo.schemas.resources import WorkerProductivityMode
from sampo.schemas.time import Time
from sampo.schemas.works import WorkUnit
from sampo.utilities.collections_util import build_index


class WorkEstimationMode(Enum):
    Pessimistic = -1,
    Realistic = 0,
    Optimistic = 1


class WorkTimeEstimator(ABC):
    """
    Implementation of time estimator of work with a given set of resources.
    """

    @abstractmethod
    def set_estimation_mode(self, use_idle: bool = True, mode: WorkEstimationMode = WorkEstimationMode.Realistic):
        ...

    @abstractmethod
    def set_productivity_mode(self, mode: WorkerProductivityMode = WorkerProductivityMode.Static):
        ...

    @abstractmethod
    def find_work_resources(self, work_name: str, work_volume: float, resource_name: list[str] | None = None) \
            -> list[WorkerReq]:
        ...

    @abstractmethod
    def estimate_time(self, work_unit: WorkUnit, worker_list: list[Worker]):
        ...


class DefaultWorkEstimator(WorkTimeEstimator):

    def __init__(self,
                 rand: Random = Random()):
        self._use_idle = True
        self._estimation_mode = WorkEstimationMode.Realistic
        self.rand = rand
        self._productivity_mode = WorkerProductivityMode.Static

    def find_work_resources(self, work_name: str, work_volume: float, resource_name: list[str] | None = None) \
            -> list[WorkerReq]:
        if resource_name is None:
            resource_name = ['driver', 'fitter', 'manager', 'handyman', 'electrician', 'engineer']
        dist = numpy.random.poisson(work_volume * 3, len(resource_name))
        return [WorkerReq(kind=name,
                          volume=work_volume * numpy.random.poisson(work_volume ** 0.5, 1)[0],
                          min_count=dist[i],
                          max_count=dist[i] * 2)
                for i, name in enumerate(resource_name)]

    def set_estimation_mode(self, use_idle: bool = True, mode: WorkEstimationMode = WorkEstimationMode.Realistic):
        self._use_idle = use_idle
        self._estimation_mode = mode

    def set_productivity_mode(self, mode: WorkerProductivityMode = WorkerProductivityMode.Static):
        self._productivity_mode = mode

    def estimate_static(self, work_unit: WorkUnit, worker_list: list[Worker]) -> Time:
        """
        Calculate summary time of task execution (without stochastic part)

        :param work_unit:
        :param worker_list:
        :return: time of task execution
        """
        # TODO Is it should be here, not in preprocessing???
        # TODO Move it to ksg_scheduling
        # workers = {w.name.replace('_res_fact', ""): w.count for w in worker_list}
        # work_time = self.estimate_time(work_unit.name.split('_stage_')[0], work_unit.volume, workers)
        # if work_time > 0:
        #     return work_time

        return self.estimate_time(work_unit, worker_list)

    def estimate_time(self, work_unit: WorkUnit, worker_list: list[Worker]) -> Time:
        if not worker_list:
            return Time(0)

        times = [Time(0)]  # if there are no requirements for the work, it is done instantly
        name2worker = build_index(worker_list, attrgetter('name'))

        for req in work_unit.worker_reqs:
            if req.min_count == 0:
                continue
            name = req.kind
            worker = name2worker.get(name, None)
            worker_count = 0 if worker is None else worker.count
            if worker_count < req.min_count:
                return Time.inf()
            productivity = DefaultWorkEstimator.get_productivity_of_worker(worker, self.rand, req.max_count,
                                                                           self._productivity_mode) / worker_count
            if productivity == 0:
                return Time.inf()
            times.append(req.volume // productivity)
        return max(max(times), Time(0))

    @staticmethod
    def get_productivity_of_worker(worker: Worker, rand: Optional[Random] = None, max_groups: int = 0,
                                   productivity_mode: WorkerProductivityMode = WorkerProductivityMode.Static):
        """
        Calculate the productivity of the Worker
        It has 2 mods: stochastic and non-stochastic, depending on the value of rand

        :param productivity_mode:
        :param max_groups:
        :param worker: the worker
        :param rand: parameter for stochastic part
        :return: productivity of received worker
        """
        return worker.get_productivity(rand, productivity_mode) * communication_coefficient(worker.count, max_groups)


def communication_coefficient(groups_count: int, max_groups: int) -> float:
    n = groups_count
    m = max_groups
    return 1 / (6 * m ** 2) * (-2 * n ** 3 + 3 * n ** 2 + (6 * m ** 2 - 1) * n)



import datetime
from typing import Tuple

import numpy as np
import pandas as pd

from sampo.schemas.graph import EdgeType
from sampo.utilities.task_name import NameMapper


def get_all_connections(graph_df: pd.DataFrame, use_mapper: bool = False, mapper: NameMapper = None) \
        -> Tuple[dict[str, list], dict[str, list]]:

    task_name_column = 'activity_name'
    if 'granular_name' in graph_df:
        task_name_column = 'granular_name'

    num_tasks = len(graph_df)
    # Get the upper triangular indices to avoid duplicate pairs
    indices = np.triu_indices(num_tasks, k=1)
    works1_ids = graph_df['activity_id'].values[indices[0]]
    works1_names = graph_df[task_name_column].values[indices[0]]

    works2_ids = graph_df['activity_id'].values[indices[1]]
    works2_names = graph_df[task_name_column].values[indices[1]]

    if use_mapper:
        works1_names = np.vectorize(mapper.get)(works1_names)
        works2_names = np.vectorize(mapper.get)(works2_names)

    return {"ids": works1_ids, "names": works1_names}, {"ids": works2_ids, "names": works2_names}


def get_delta_between_dates(first: str, second: str) -> int:
    return max((datetime.date(int(first.split('-')[0]), int(first.split('-')[1]), int(first.split('-')[2])) -
                datetime.date(int(second.split('-')[0]), int(second.split('-')[1]), int(second.split('-')[2]))).days, 1)


def find_min_without_outliers(lst: list[float]) -> float:
    return round(min([x for x in lst if x >= np.mean(lst) - 3 * np.std(lst)]), 2)


def gather_links_types_statistics(s1: str, f1: str, s2: str, f2: str) \
        -> Tuple[int, int, int, list, list, int, list, list, int, list, list, int, list, list]:
    """
    Count statistics on the occurrence of different mutual arrangement of tasks

    :param s1: start of first work
    :param f1: finish of first work
    :param s2: start of second work
    :param f2: finish of second work
    :return: Statistics on the occurrence of different mutual arrangement of tasks
    """

    fs12, fs21, ss12, ss21 = 0, 0, 0, 0
    ss12_lags, ss12_percent_lags, ss21_lags, ss21_percent_lags = [], [], [], []

    ffs12, ffs21 = 0, 0
    ffs12_lags, ffs12_percent_lags, ffs21_lags, ffs21_percent_lags = [], [], [], []

    if s1 == s2 and f1 == f2:
        ffs12 += 1
        ffs12_percent_lags.append(0.01)
        ffs12_lags.append(0.01)
    if f2 <= s1:
        fs21 += 1
    else:
        if s2 >= f1:
            fs12 += 1
        else:
            if s2 >= s1:
                if f2 >= f1:
                    ffs12 += 1
                    if get_delta_between_dates(f1, s1) != 0:
                        ffs12_percent_lags.append(get_delta_between_dates(s2, s1) / get_delta_between_dates(f1, s1))
                    else:
                        ffs12_percent_lags.append(0)
                    ffs12_lags.append(get_delta_between_dates(s2, s1))
                else:
                    ss12 += 1
                    if get_delta_between_dates(f1, s1) != 0:
                        ss12_percent_lags.append(get_delta_between_dates(s2, s1) / get_delta_between_dates(f1, s1))
                    else:
                        ss12_percent_lags.append(0)
                    ss12_lags.append(get_delta_between_dates(s2, s1))
            else:
                if f2 <= f1:
                    ffs21 += 1
                    if get_delta_between_dates(f2, s2) != 0:
                        ffs21_percent_lags.append(get_delta_between_dates(s1, s2) / get_delta_between_dates(f2, s2))
                    else:
                        ffs21_percent_lags.append(0)
                    ffs21_lags.append(get_delta_between_dates(s1, s2))
                else:
                    ss21 += 1
                    if get_delta_between_dates(f2, s2) != 0:
                        ss21_percent_lags.append(get_delta_between_dates(s1, s2) / get_delta_between_dates(f2, s2))
                    else:
                        ss21_percent_lags.append(0)
                    ss21_lags.append(get_delta_between_dates(s1, s2))
    return fs12, fs21, ss12, ss12_lags, ss12_percent_lags, ss21, ss21_lags, ss21_percent_lags, ffs12, ffs12_lags, ffs12_percent_lags, \
        ffs21, ffs21_lags, ffs21_percent_lags


def get_all_seq_statistic(history_data, graph_df, use_model_name=False, mapper=None):
    df_grouped = history_data.copy()

    if use_model_name:
        column_name = 'model_name'
    else:
        column_name = 'granular_smr_name'

    df_grouped = df_grouped.groupby('upper_works')[column_name].apply(list).reset_index(name="Works")
    works1, works2 = get_all_connections(graph_df, use_model_name, mapper)

    # Declare structure with updated connections

    tasks_names = list(zip(works1['names'], works2['names']))
    tasks_ids = list(zip(works1['ids'], works2['ids']))

    predecessors_info_dict = {w_id: [] for w_id in graph_df['activity_id']}

    if len(tasks_names) != 0:
        for i in range(len(tasks_names)):
            w1, w2 = tasks_names[i]
            w1_id, w2_id = tasks_ids[i]

            if w1 != w2:
                fs12, fs21, ss12, ss21 = 0, 0, 0, 0
                ss12_lags, ss12_percent_lags, ss21_lags, ss21_percent_lags = [], [], [], []

                count = 0

                ffs12, ffs21 = 0, 0
                ffs12_lags, ffs12_percent_lags, ffs21_lags, ffs21_percent_lags = [], [], [], []

                for i, work_list in df_grouped.iterrows():
                    # Looking to see if this pair of works occurred within the same site in the historical data
                    if w1 in work_list['Works'] and w2 in work_list['Works']:
                        ind1 = history_data.loc[(history_data['upper_works'] == work_list['upper_works']) &
                                                (history_data['granular_smr_name'] == w1)]
                        ind2 = history_data.loc[(history_data['upper_works'] == work_list['upper_works']) &
                                                (history_data['granular_smr_name'] == w2)]

                        ind1_sorted = ind1.sort_values(by=['first_day', 'last_day']).reset_index(drop=True)
                        ind2_sorted = ind2.sort_values(by=['first_day', 'last_day']).reset_index(drop=True)

                        for l in range(min(len(ind1_sorted), len(ind2_sorted))):
                            s1, f1 = ind1_sorted.loc[l, 'first_day'], ind1_sorted.loc[l, 'last_day']

                            s2, f2 = ind2_sorted.loc[l, 'first_day'], ind2_sorted.loc[l, 'last_day']

                            if not any([type(x) == float for x in [s1, s2, f1, f2]]):
                                tasks_fs12, tasks_fs21, tasks_ss12, tasks_ss12_lags, tasks_ss12_percent_lags, tasks_ss21, tasks_ss21_lags, \
                                    tasks_ss21_percent_lags, tasks_ffs12, tasks_ffs12_lags, tasks_ffs12_percent_lags, tasks_ffs21, tasks_ffs21_lags, tasks_ffs21_percent_lags = gather_links_types_statistics(
                                    s1, f1, s2, f2)

                                count += 1

                                fs12 += tasks_fs12
                                fs21 += tasks_fs21

                                ss12 += tasks_ss12
                                ss12_lags.extend(tasks_ss12_lags)
                                ss12_percent_lags.extend(tasks_ss12_percent_lags)
                                ss21 += tasks_ss21
                                ss21_lags.extend(tasks_ss21_lags)
                                ss21_percent_lags.extend(tasks_ss21_percent_lags)

                                ffs12 += tasks_ffs12
                                ffs12_lags.extend(tasks_ffs12_lags)
                                ffs12_percent_lags.extend(tasks_ffs12_percent_lags)
                                ffs21 += tasks_ffs21
                                ffs21_lags.extend(tasks_ffs21_lags)
                                ffs21_percent_lags.extend(tasks_ffs21_percent_lags)

                if fs12 + ffs12 + ss12 >= fs21 + ffs21 + ss21:
                    order_con = 1
                    fs = fs12
                    ffs = ffs12
                    ss = ss12
                else:
                    order_con = 2
                    fs = fs21
                    ffs = ffs21
                    ss = ss21

                if max([fs, ss, ffs]) != 0:
                    if fs > ss:
                        if ffs > 0:
                            if order_con == 1:
                                predecessors_info_dict[w2_id].append([w1_id, 'FFS',
                                                                      find_min_without_outliers(ffs12_percent_lags),
                                                                      count])
                            else:
                                predecessors_info_dict[w1_id].append([w2_id, 'FFS',
                                                                      find_min_without_outliers(ffs21_percent_lags),
                                                                      count])
                        else:
                            if order_con == 1:
                                predecessors_info_dict[w2_id].append([w1_id, 'FS', -1, count])
                            else:
                                predecessors_info_dict[w1_id].append([w2_id, 'FS', -1, count])
                    elif ss > ffs:
                        if order_con == 1:
                                predecessors_info_dict[w2_id].append([w1_id, 'SS',
                                                                      find_min_without_outliers(ss12_percent_lags),
                                                                      count])
                        else:
                            predecessors_info_dict[w1_id].append([w2_id, 'SS',
                                                                  find_min_without_outliers(ss21_percent_lags), count])
                    else:
                        if order_con == 1:
                                predecessors_info_dict[w2_id].append([w1_id, 'FFS',
                                                                      find_min_without_outliers(ffs12_percent_lags),
                                                                      count])
                        else:
                            predecessors_info_dict[w1_id].append([w2_id, 'FFS',
                                                                  find_min_without_outliers(ffs21_percent_lags), count])

    return predecessors_info_dict


def set_connections_info(graph_df: pd.DataFrame,
                         history_data: pd.DataFrame,
                         use_model_name: bool = False,
                         mapper=None,
                         change_connections_info: bool = False,
                         expert_connections_info: bool = False) \
        -> pd.DataFrame:
    """
    Restore tasks' connection based on history data

    :param: change_connections_info - whether existing connections should be modified based on connection history data
    :param: expert_connections_info - whether existing connections should not be modified based on connection history data
    :return: repaired DataFrame
    """
    tasks_df = graph_df.copy().set_index('activity_id', drop=False)
    connections_dict = get_all_seq_statistic(history_data, graph_df, use_model_name, mapper)
    # connections_dict = {'25809398': [], '25809830': [], '25809831': [['25809830', 'FFS', 0.01, 185], ['25809856', 'FS', -1, 3]], '25809833': [['25809830', 'FFS', 0.01, 276], ['25809831', 'FFS', 0.01, 1646]], '25813507': [['25809830', 'FFS', 0.01, 60], ['25809831', 'FFS', 0.01, 131], ['25809833', 'FFS', 0.01, 423]], '25809836': [['25809830', 'FFS', 0.01, 113], ['25809831', 'FFS', 0.01, 907], ['25809833', 'FFS', 0.01, 1310], ['25813507', 'FFS', 0.01, 278]], '25809832': [['25809830', 'FS', -1, 3], ['25809833', 'FS', -1, 3], ['25809836', 'FFS', 0.01, 5], ['25809839', 'SS', 0.05, 6], ['25809852', 'FS', -1, 2], ['25809854', 'FS', -1, 2]], '25809837': [['25809831', 'FS', -1, 2], ['25809836', 'SS', 0.75, 2], ['25809835', 'FS', -1, 2]], '25809838': [['25809830', 'FS', -1, 3], ['25809831', 'FFS', 0.01, 259], ['25809833', 'FFS', 0.01, 379], ['25813507', 'FFS', 0.01, 111], ['25809836', 'FFS', 0.01, 500]], '25809839': [['25809830', 'FFS', 0.52, 25], ['25809831', 'FFS', 0.01, 345], ['25809833', 'FFS', 0.01, 419], ['25813507', 'FFS', 0.33, 61], ['25809836', 'FFS', 0.01, 497], ['25809838', 'FFS', 0.01, 372], ['25809835', 'FS', -1, 2], ['25809847', 'FFS', 0.27, 10], ['25809850', 'FS', -1, 1], ['25809852', 'FFS', 0.05, 97], ['25809848', 'FFS', 0.27, 10], ['25809853', 'FS', -1, 1], ['25809854', 'FFS', 0.05, 97]], '25809834': [['25809830', 'FS', -1, 5], ['25809831', 'FS', -1, 6], ['25809833', 'FS', -1, 7], ['25809836', 'FFS', 0.38, 13], ['25809838', 'FFS', 0.5, 8], ['25809839', 'SS', 0.68, 4], ['25809835', 'FS', -1, 18], ['25809840', 'FS', -1, 18], ['25809841', 'FFS', 0.86, 12]], '25809835': [['25809831', 'FFS', 0.04, 6], ['25809833', 'FFS', 0.04, 6], ['25809836', 'SS', 0.19, 8], ['25809838', 'SS', 0.5, 5], ['25809841', 'SS', 0.76, 1], ['25809852', 'FS', -1, 2], ['25809854', 'FS', -1, 2]], '25809842': [['25809830', 'FFS', 0.01, 6], ['25809831', 'FFS', 0.01, 2], ['25809833', 'FFS', 0.01, 8], ['25813507', 'FFS', 0.25, 1], ['25809836', 'FFS', 0.01, 9], ['25809832', 'FFS', 0.01, 1], ['25809839', 'FFS', 0.01, 2], ['25809840', 'SS', 0.79, 1], ['25809841', 'FS', -1, 11]], '25809843': [['25809832', 'FFS', 0.01, 23], ['25809839', 'FS', -1, 1], ['25809842', 'FFS', 0.06, 1], ['25809841', 'FS', -1, 4]], '25809844': [['25809832', 'FFS', 0.01, 4], ['25809843', 'SS', 0.03, 4], ['25809841', 'FS', -1, 2]], '25809840': [['25809830', 'FS', -1, 2], ['25809831', 'FFS', 0.01, 40], ['25809833', 'FFS', 0.01, 43], ['25809836', 'FFS', 0.17, 45], ['25809838', 'FS', -1, 8], ['25809839', 'FS', -1, 13], ['25809835', 'FFS', 0.01, 24], ['25809841', 'FS', -1, 19], ['25809847', 'FS', -1, 7], ['25809850', 'FFS', 0.12, 1], ['25809848', 'FS', -1, 7], ['25809853', 'FFS', 0.12, 1]], '25809841': [['25809830', 'FFS', 0.01, 202], ['25809831', 'FFS', 0.01, 334], ['25809833', 'FFS', 0.01, 446], ['25813507', 'FFS', 0.01, 105], ['25809836', 'FFS', 0.01, 358], ['25809832', 'SS', 0.32, 7], ['25809838', 'FFS', 0.01, 103], ['25809839', 'FFS', 0.01, 190], ['25809847', 'FS', -1, 17], ['25809850', 'FS', -1, 2], ['25809852', 'FS', -1, 2], ['25809848', 'FS', -1, 17], ['25809853', 'FS', -1, 2], ['25809854', 'FS', -1, 2], ['25809856', 'FS', -1, 1]], '25809399': [], '25809400': [], '25809847': [['25809830', 'FS', -1, 1], ['25809831', 'FFS', 0.01, 81], ['25809833', 'FFS', 0.01, 85], ['25809836', 'FFS', 0.01, 133], ['25809838', 'FFS', 0.01, 54], ['25809834', 'FS', -1, 2], ['25809850', 'FFS', 0.03, 68], ['25809853', 'FFS', 0.03, 68], ['25809856', 'FS', -1, 2]], '25809850': [['25809831', 'FS', -1, 3], ['25809833', 'FS', -1, 3], ['25809836', 'FFS', 0.67, 8], ['25809838', 'FFS', 0.03, 44], ['25809856', 'FS', -1, 3]], '25809852': [['25809831', 'FFS', 0.68, 3], ['25809833', 'FFS', 0.68, 3], ['25809836', 'FFS', 0.46, 6], ['25809838', 'SS', 0.01, 2], ['25809847', 'FFS', 0.01, 9], ['25809850', 'FFS', 0.01, 265], ['25809857', 'FS', -1, 1], ['25809855', 'FFS', 0.16, 20]], '25809848': [['25809830', 'FS', -1, 1], ['25809831', 'FFS', 0.01, 81], ['25809833', 'FFS', 0.01, 85], ['25809836', 'FFS', 0.01, 133], ['25809838', 'FFS', 0.01, 54], ['25809834', 'FS', -1, 2], ['25809850', 'FFS', 0.01, 68], ['25809852', 'FFS', 0.01, 9], ['25809853', 'FFS', 0.03, 68], ['25809856', 'FS', -1, 2]], '25809853': [['25809831', 'FS', -1, 3], ['25809833', 'FS', -1, 3], ['25809836', 'FFS', 0.67, 8], ['25809838', 'FFS', 0.03, 44], ['25809852', 'FFS', 0.01, 265], ['25809856', 'FS', -1, 3]], '25809854': [['25809831', 'FFS', 0.68, 3], ['25809833', 'FFS', 0.68, 3], ['25809836', 'FFS', 0.46, 6], ['25809838', 'SS', 0.01, 2], ['25809847', 'FFS', 0.01, 9], ['25809850', 'FFS', 0.01, 265], ['25809848', 'FFS', 0.01, 9], ['25809853', 'FFS', 0.01, 265], ['25809857', 'FS', -1, 1], ['25809855', 'FFS', 0.16, 20]], '25809856': [['25809833', 'FS', -1, 43], ['25813507', 'FFS', 0.91, 33], ['25809836', 'FFS', 0.47, 41], ['25809838', 'FFS', 0.01, 7], ['25809839', 'FFS', 0.01, 12], ['25809835', 'FS', -1, 1]], '25809401': [], '25809857': [['25809855', 'FS', -1, 1]], '25809858': [['25809833', 'FS', -1, 1], ['25809836', 'FS', -1, 1], ['25809838', 'FS', -1, 1], ['25809839', 'FFS', 0.01, 56], ['25809841', 'SS', 0.36, 4], ['25809847', 'FFS', 0.01, 17], ['25809850', 'FFS', 0.93, 91], ['25809852', 'FS', -1, 89], ['25809848', 'FFS', 0.01, 17], ['25809853', 'FFS', 0.93, 91], ['25809854', 'FS', -1, 89], ['25809856', 'SS', 0.64, 1], ['25809857', 'FS', -1, 1]], '25809855': [['25809850', 'FFS', 0.12, 4], ['25809853', 'FFS', 0.12, 4]]}

    predecessors_ids_lst, predecessors_types_lst, predecessors_lags_lst, predecessors_counts_lst = [], [], [], []

    for task_id, pred_info_lst in connections_dict.items():
        if len(pred_info_lst) > 0:
            pred_ids_lst, pred_types_lst, pred_lags_lst, pred_counts_lst = map(list, zip(*pred_info_lst))
        else:
            pred_ids_lst, pred_types_lst, pred_lags_lst, pred_counts_lst = ['-1'], ['-1'], [-1], [0]

        if str(task_id) in tasks_df.index:
            if tasks_df.loc[str(task_id), 'predecessor_ids'] != ['-1']:
                if expert_connections_info:
                    predecessors_ids_lst.append(tasks_df.loc[str(task_id), 'predecessor_ids'])
                    predecessors_types_lst.append(tasks_df.loc[str(task_id), 'connection_types'])
                    predecessors_lags_lst.append(tasks_df.loc[str(task_id), 'lags'])
                    predecessors_counts_lst.append(pred_counts_lst)
                    continue
                if change_connections_info:
                    predecessors_ids_lst.append(tasks_df.loc[str(task_id), 'predecessor_ids'])
                else:
                    predecessors_ids_lst.append(pred_ids_lst)
            else:
                predecessors_ids_lst.append(pred_ids_lst)
        else:
            predecessors_ids_lst.append(pred_ids_lst)

        predecessors_types_lst.append(pred_types_lst)
        predecessors_lags_lst.append(pred_lags_lst)
        predecessors_counts_lst.append(pred_counts_lst)
        while len(predecessors_types_lst[-1]) != len(predecessors_ids_lst[-1]):
            predecessors_types_lst[-1].append('FS')
            predecessors_lags_lst[-1].append(-1)
            predecessors_counts_lst[-1].append(0)

    # Convert strings to arrays
    tasks_df['predecessor_ids'] = predecessors_ids_lst
    tasks_df['connection_types'] = predecessors_types_lst
    tasks_df['lags'] = predecessors_lags_lst
    tasks_df['counts'] = predecessors_counts_lst

    tasks_df['connection_types'] = tasks_df['connection_types'].apply(
        lambda x: [EdgeType(elem) if elem != '-1' else EdgeType.FinishStart for elem in x]
    )

    return tasks_df



from datetime import timedelta
from typing import Optional

import pandas as pd
import plotly.express as px

from sampo.utilities.visualization.base import VisualizationMode, visualize


def schedule_gant_chart_fig(schedule_dataframe: pd.DataFrame,
                            visualization: VisualizationMode,
                            remove_service_tasks: bool = False,
                            fig_file_name: Optional[str] = None):
    """
    Creates and saves a gant chart of the scheduled tasks to the specified path.

    :param fig_file_name:
    :param visualization:
    :param remove_service_tasks:
    :param schedule_dataframe: Pandas DataFrame with the information about schedule
    """
    if remove_service_tasks:
        schedule_dataframe = schedule_dataframe.loc[~schedule_dataframe.loc[:, 'task_name'].str.contains('марке')]
        schedule_dataframe = schedule_dataframe.loc[schedule_dataframe.loc[:, 'volume'] >= 0.1]

    schedule_dataframe = schedule_dataframe.rename({'workers': 'workers_dict'}, axis=1)
    schedule_dataframe.loc[:, 'workers'] = schedule_dataframe.loc[:, 'workers_dict']\
        .apply(lambda x: x.replace(", '", ", <br>'"))

    schedule_start = schedule_dataframe.loc[:, 'start'].min()
    schedule_finish = schedule_dataframe.loc[:, 'finish'].max()
    visualization_start_delta = timedelta(days=2)
    visualization_finish_delta = timedelta(days=(schedule_finish - schedule_start).days // 3)

    def create_zone_row(i, zone_names, zone) -> dict:
        return {'idx': i,
                'contractor': 'Access cards',
                'cost': 0,
                'volume': 0,
                'duration': 0,
                'measurement': 'unit',
                'workers_dict': '',
                'workers': '',
                'task_name_mapped': zone_names,
                'task_name': '',
                'zone_information': '',
                'start': timedelta(int(zone.start_time)) + schedule_start,
                'finish': timedelta(int(zone.end_time)) + schedule_start}

    sworks = schedule_dataframe['scheduled_work_object'].copy()
    idx = schedule_dataframe['idx'].copy()

    def get_zone_usage_info(swork) -> str:
        return '<br>' + '<br>'.join([f'{zone.name}: {zone.to_status}' for zone in swork.zones_pre])

    schedule_dataframe['zone_information'] = sworks.apply(get_zone_usage_info)

    access_cards = []

    # create zone information
    for i, swork in zip(idx, sworks):
        zone_names = '<br>' + '<br>'.join([zone.name for zone in swork.zones_pre])
        for zone in swork.zones_pre:
            access_cards.append(create_zone_row(i, zone_names, zone))
        zone_names = '<br>' + '<br>'.join([zone.name for zone in swork.zones_post])
        for zone in swork.zones_post:
            access_cards.append(create_zone_row(i, zone_names, zone))

    schedule_dataframe = pd.concat([schedule_dataframe, pd.DataFrame.from_records(access_cards)])

    schedule_dataframe['color'] = schedule_dataframe[['task_name', 'contractor']] \
        .apply(lambda r: 'Defect' if ':' in r['task_name'] else r['contractor'], axis=1)
    schedule_dataframe['idx'] = (schedule_dataframe[['idx', 'task_name']]
                                 .apply(lambda r: schedule_dataframe[schedule_dataframe['task_name'] ==
                                                                     r['task_name'].split(':')[0]]['idx'].iloc[0]
                                 if ':' in r['task_name'] else r['idx'], axis=1))

    # add one time unit to the end should remove hole within the immediately close tasks
    schedule_dataframe['vis_finish'] = schedule_dataframe[['start', 'finish', 'duration']] \
        .apply(lambda r: r['finish'] + timedelta(1) if r['duration'] > 0 else r['finish'], axis=1)
    schedule_dataframe['vis_start'] = schedule_dataframe['start']
    schedule_dataframe['finish'] = schedule_dataframe['finish'].apply(lambda x: x.strftime('%e %b %Y'))
    schedule_dataframe['start'] = schedule_dataframe['start'].apply(lambda x: x.strftime('%e %b %Y'))

    fig = px.timeline(schedule_dataframe, x_start='vis_start', x_end='vis_finish', y='idx', hover_name='task_name',
                      color=schedule_dataframe.loc[:, 'color'],
                      hover_data={'vis_start': False,
                                  'vis_finish': False,
                                  'start': True,
                                  'finish': True,
                                  'task_name_mapped': True,
                                  'cost': True,
                                  'volume': True,
                                  'measurement': True,
                                  'workers': True,
                                  'zone_information': True},
                      title=f"{'Project tasks - Gant chart'}",
                      category_orders={'idx': list(schedule_dataframe.idx)},
                      text='task_name')

    fig.update_traces(textposition='outside')

    fig.update_yaxes(showticklabels=False, title_text='Project tasks',
                     type='category')

    fig.update_xaxes(type='date',
                     range=[schedule_start - visualization_start_delta,
                            schedule_finish + visualization_finish_delta],
                     title_text='Date')

    fig.update_layout(autosize=True, font_size=12)
    # fig.update_layout(height=1000)

    return visualize(fig, mode=visualization, file_name=fig_file_name)



from sampo.scheduler.base import Scheduler, SchedulerType
from sampo.scheduler.genetic import GeneticScheduler
from sampo.scheduler.generic import GenericScheduler
from sampo.scheduler.heft import HEFTScheduler, HEFTBetweenScheduler
from sampo.scheduler.topological import TopologicalScheduler
from sampo.scheduler.native_wrapper import NativeWrapper
